# coding: utf-8
"""deploy and connect to running serving apps"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List

import supervisely.io.env as sly_env
import supervisely.io.env as env
from supervisely.sly_logger import logger

if TYPE_CHECKING:
    from supervisely.api.api import Api
    from supervisely.nn.experiments import ExperimentInfo
    from supervisely.nn.model.model_api import ModelAPI


class NeuralNetworkApi:
    """
    API to interact with neural networks in Supervisely.
    It provides methods to deploy and connect to models for running inference.
    """

    def __init__(self, api: "Api"):
        from supervisely.api.nn.deploy_api import DeployApi
        from supervisely.api.nn.ecosystem_models_api import EcosystemModelsApi
        from supervisely.api.nn.train_api import TrainApi

        self._api = api
        self._deploy_api = DeployApi(api)
        self.ecosystem_models_api = EcosystemModelsApi(api)
        self.train = TrainApi(api)

    def deploy(
        self,
        model: str,
        device: str = None,
        runtime: str = None,
        workspace_id: int = None,
        agent_id: int = None,
        **kwargs,
    ) -> ModelAPI:
        """
        Deploy a pretrained model or a custom model checkpoint in Supervisely platform.
        This method will start a new Serving App in Supervisely, deploy a given model, and return a `ModelAPI` object for running predictions and managing the model.
        - To deploy a pretrained model, pass the model name in the format `framework/model_name` (e.g., "RT-DETRv2/RT-DETRv2-M").
        - To deploy a custom model, pass the path to the model checkpoint in team files (e.g., "/experiments/1089_RT-DETRv2/checkpoints/best.pt").

        :param model: Either a path to a model checkpoint in team files or model name in format `framework/model_name` (e.g., "RT-DETRv2/RT-DETRv2-M").
        :type model: str
        :param device: Device to run the model on (e.g., "cuda:0" or "cpu"). If not specified, will automatically use GPU device if available, otherwise CPU will be used.
        :type device: Optional[str]
        :param runtime: If specified, the model will be converted to the given format (e.g., "onnx", "tensorrt") and will be deployed in the corresponding accelerated runtime. This option is used for pretrained models. For custom models, the runtime will be defined automatically based on the model checkpoint.
        :type runtime: Optional[str]
        :param workspace_id: Workspace ID, if None, will be got from env.
        :type workspace_id: Optional[int]
        :param agent_id: Agent ID, if not present will be defined automatically.
        :type agent_id: Optional[int]
        :param kwargs: Additional parameters for deployment.
        :return: A :class:`ModelAPI` object for the deployed model.
        :rtype: ModelAPI
        :Usage example:
            .. code-block:: python

                import supervisely as sly

                api = sly.Api()
                model = api.nn.deploy(model="RT-DETRv2/RT-DETRv2-M")
        """

        from supervisely.nn.model.model_api import ModelAPI

        checkpoint = None
        pretrained = None
        team_id = None
        if workspace_id is None:
            workspace_id = sly_env.workspace_id(raise_not_found=False)
        if workspace_id is None:
            raise ValueError(
                "Workspace ID is not specified. Please, provide it in the function call, or set 'WORKSPACE_ID' variable in the environment."
            )
        if team_id is None:
            workspace_info = self._api.workspace.get_info_by_id(workspace_id)
            team_id = workspace_info.team_id
        if agent_id is None:
            agent_id = self._deploy_api._find_agent(team_id)

        if model.startswith("/"):
            checkpoint = model
        else:
            found_team_id = self._deploy_api._find_team_by_path(
                f"/{model}", team_id=team_id, raise_not_found=False
            )
            if found_team_id is not None:
                checkpoint = f"/{model}"
                team_id = found_team_id
                logger.debug(f"Found checkpoint in team {team_id}")
            else:
                pretrained = model

        if checkpoint is not None:
            logger.debug(f"Deploying model by checkpoint: {checkpoint}")
            task_info = self._deploy_api.deploy_custom_model_by_checkpoint(
                checkpoint=checkpoint,
                device=device,
                runtime=runtime,
                team_id=team_id,
                workspace_id=workspace_id,
                agent_id=agent_id,
                **kwargs,
            )
        else:
            framework, model_name = pretrained.split("/", 1)
            logger.debug(
                f"Deploying pretrained model. Framework: {framework}, Model name: {model_name}"
            )
            task_info = self._deploy_api.deploy_pretrained_model(
                framework=framework,
                model_name=model_name,
                device=device,
                runtime=runtime,
                workspace_id=workspace_id,
                agent_id=agent_id,
                **kwargs,
            )
        return ModelAPI(self._api, task_id=task_info["id"])

    def list_deployed_models(
        self,
        model: str = None,
        framework: str = None,
        task_type: str = None,
        team_id: int = None,
        workspace_id: int = None,
    ) -> List[Dict]:
        """
        Returns a list of deployed models in the Supervisely platform.
        The list can be filtered by model name, framework, task type, team ID, and workspace ID.

        :param model: Model name or checkpoint path to filter the results. If None, all models will be returned.
        :type model: Optional[str]
        :param framework: Framework name to filter the results. If None, all frameworks will be returned.
        :type framework: Optional[str]
        :param task_type: CV Task to filter the results, e.g., "object detection", "instance segmentation", etc. If None, all task types will be returned.
        :type task_type: Optional[str]
        :param team_id: Team ID to filter the results. If None, the team ID from the environment will be used.
        :type team_id: Optional[int]
        :param workspace_id: Workspace ID to filter the results. If None, the workspace ID from the environment will be used.
        :type workspace_id: Optional[int]
        :return: A list of dictionaries containing information about the deployed models.
        :rtype: List[Dict]
        :Usage example:
            .. code-block:: python

                import supervisely as sly

                api = sly.Api()
                deployed_models = api.nn.list_deployed_models(framework="RT-DETRv2")
        """
        # 1. Define apps
        categories = ["serve"]
        if framework is not None:
            categories.append(f"framework:{framework}")
        serve_apps = self._api.app.get_list_ecosystem_modules(
            categories=categories, categories_operation="and"
        )
        if not serve_apps:
            return []
        serve_apps_module_ids = {app["id"] for app in serve_apps}
        # 2. Get tasks infos
        if workspace_id is not None:
            workspaces = [workspace_id]
        elif team_id is not None:
            workspaces = self._api.workspace.get_list(team_id)
            workspaces = [workspace.id for workspace in workspaces]
        else:
            workspace_id = env.workspace_id(raise_not_found=False)
            if workspace_id is None:
                team_id = env.team_id(raise_not_found=False)
                if team_id is None:
                    raise ValueError(
                        "Workspace ID and Team ID are not specified and cannot be found in the environment."
                    )
                workspaces = self._api.workspace.get_list(team_id)
                workspaces = [workspace.id for workspace in workspaces]
            else:
                workspaces = [workspace_id]

        all_tasks = []
        for workspace_id in workspaces:
            all_tasks.extend(
                self._api.task.get_list(
                    workspace_id=workspace_id,
                    filters=[
                        {
                            "field": "status",
                            "operator": "in",
                            "value": [str(self._api.task.Status.STARTED)],
                        }
                    ],
                )
            )
        all_tasks = [
            task for task in all_tasks if task["meta"]["app"]["moduleId"] in serve_apps_module_ids
        ]
        # get deploy infos and filter results
        result = []
        for task in all_tasks:
            try:
                deploy_info = self._deploy_api.get_deploy_info(task["id"])
            except Exception as e:
                logger.warning(
                    f"Failed to get deploy info for task {task['id']}: {e}",
                    exc_info=True,
                )
                continue
            if model is not None:
                checkpoint = deploy_info["checkpoint_name"]
                deployed_model = deploy_info["model_name"]
                if checkpoint != model and not model.endswith(deployed_model):
                    continue
            if task_type is not None:
                if deploy_info["task_type"] != task_type:
                    continue
            result.append(
                {
                    "task_info": task,
                    "model_info": deploy_info,
                }
            )
        return result

    def get_experiment_info(self, task_id: int) -> ExperimentInfo:
        """
        Returns the experiment info of a finished training task by its task_id.

        :param task_id: the task_id of a finished training task in the Supervisely platform.
        :type task_id: int
        :return: an :class:`ExperimentInfo` object with information about the training, model, and results.
        :rtype: ExperimentInfo
        """
        from supervisely.nn.experiments import ExperimentInfo

        task_info = self._api.task.get_info_by_id(task_id)
        if task_info is None:
            raise ValueError(f"Task with ID '{task_id}' not found")
        try:
            data = task_info["meta"]["output"]["experiment"]["data"]
            return ExperimentInfo(**data)
        except KeyError:
            raise ValueError("Task output does not contain experiment data")

    def connect(
        self,
        task_id: int,
    ) -> ModelAPI:
        """
        Connect to a running Serving App by its `task_id`. This allows you to make predictions and control the model state via API.

        :param task_id: the task_id of a running Serving App session in the Supervisely platform.
        :type task_id: int
        :return: a :class:`ModelAPI` object
        :rtype: ModelAPI
        """
        from supervisely.nn.model.model_api import ModelAPI

        return ModelAPI(self._api, task_id)



import json
import pickle
import zipfile
from pathlib import Path
from typing import Dict, List, Optional

import numpy as np
import pandas as pd

from supervisely.annotation.annotation import ProjectMeta
from supervisely.api.api import Api
from supervisely.api.dataset_api import DatasetInfo
from supervisely.api.module_api import ApiField
from supervisely.api.project_api import ProjectInfo
from supervisely.app.widgets import SlyTqdm
from supervisely.io.env import team_id
from supervisely.io.fs import dir_empty, mkdir
from supervisely.io.json import load_json_file
from supervisely.nn.benchmark.object_detection.metric_provider import MetricProvider
from supervisely.nn.benchmark.visualization.vis_click_data import ClickData, IdMapper
from supervisely.sly_logger import logger
from supervisely.task.progress import tqdm_sly

# class ImageComparisonData:
#     def __init__(
#         self,
#         gt_image_info: ImageInfo = None,
#         pred_image_info: ImageInfo = None,
#         diff_image_info: ImageInfo = None,
#         gt_annotation: Annotation = None,
#         pred_annotation: Annotation = None,
#         diff_annotation: Annotation = None,
#     ):
#         self.gt_image_info = gt_image_info
#         self.pred_image_info = pred_image_info
#         self.diff_image_info = diff_image_info
#         self.gt_annotation = gt_annotation
#         self.pred_annotation = pred_annotation
#         self.diff_annotation = diff_annotation


class EvalResult:

    def __init__(
        self,
        eval_dir: str,
        workdir: str,
        api: Api,
        progress: Optional[SlyTqdm] = None,
    ):
        from pycocotools.coco import COCO  # pylint: disable=import-error

        self.eval_dir = eval_dir
        self.report_path = Path(eval_dir, "visualizations", "template.vue").as_posix()
        self.workdir = workdir
        self.api = api
        self.team_id = team_id()
        self.local_dir = str(Path(self.workdir, self.eval_dir.lstrip("/")))
        self.progress = progress or tqdm_sly

        self.coco_gt: COCO = None
        self.coco_dt: COCO = None
        self.inference_info: Dict = None
        self.speedtest_info: Dict = None
        self.eval_data: Dict = None
        self.mp: MetricProvider = None
        self.df_score_profile: pd.DataFrame = None
        self.dfsp_down: pd.DataFrame = None
        self.f1_optimal_conf: float = None
        self.click_data: ClickData = None
        # self.comparison_data: Dict[int, ImageComparisonData] = {}
        self.color = None

        self._gt_project_info = None
        self._gt_project_meta = None
        self._gt_dataset_infos = None
        self._dt_project_id = None
        self._dt_project_meta = None

        self._load_eval_data()
        self._read_eval_data()

    @property
    def cv_task(self):
        return self.inference_info.get("task_type")

    @property
    def name(self) -> str:
        model_name = self.inference_info.get("model_name", self.eval_dir)
        return self.inference_info.get("deploy_params", {}).get("checkpoint_name", model_name)

    @property
    def model_name(self) -> str:
        if not self.name:
            return
        if len(self.name) > 20:
            return self.name[:9] + "..." + self.name[-6:]
        return self.name

    @property
    def gt_project_id(self) -> int:
        return self.inference_info.get("gt_project_id")

    @property
    def gt_project_info(self) -> ProjectInfo:
        if self._gt_project_info is None:
            gt_project_id = self.inference_info.get("gt_project_id")
            self._gt_project_info = self.api.project.get_info_by_id(gt_project_id)
        return self._gt_project_info

    @property
    def gt_project_meta(self) -> ProjectMeta:
        if self._gt_project_meta is None:
            self._gt_project_meta = ProjectMeta.from_json(
                self.api.project.get_meta(self.gt_project_id)
            )
        return self._gt_project_meta

    @property
    def gt_dataset_ids(self) -> List[int]:
        return self.inference_info.get("gt_dataset_ids", None)

    @property
    def gt_dataset_infos(self) -> List[DatasetInfo]:
        if self._gt_dataset_infos is None:
            filters = None
            if self.gt_dataset_ids is not None:
                filters = [
                    {
                        ApiField.FIELD: ApiField.ID,
                        ApiField.OPERATOR: "in",
                        ApiField.VALUE: self.gt_dataset_ids,
                    }
                ]
            self._gt_dataset_infos = self.api.dataset.get_list(
                self.gt_project_id,
                filters=filters,
                recursive=True,
            )
        return self._gt_dataset_infos

    @property
    def dt_project_id(self):
        if self._dt_project_id is None:
            self._dt_project_id = self.inference_info.get("dt_project_id")
        return self._dt_project_id

    @property
    def dt_project_meta(self):
        if self._dt_project_meta is None:
            self._dt_project_meta = ProjectMeta.from_json(
                self.api.project.get_meta(self.dt_project_id)
            )
        return self._dt_project_meta

    @property
    def train_info(self):
        return self.inference_info.get("train_info", None)  # TODO: check

    @property
    def gt_images_ids(self):
        return self.inference_info.get("gt_images_ids", None)  # TODO: check

    @property
    def classes_whitelist(self):
        return self.inference_info.get("inference_settings", {}).get("classes", [])  # TODO: check

    def _load_eval_data(self):
        if not dir_empty(self.local_dir):
            return
        if not self.api.storage.dir_exists(self.team_id, self.eval_dir):
            raise ValueError(f"Directory {self.eval_dir} not found in storage.")
        mkdir(self.local_dir)
        dir_name = Path(self.eval_dir).name
        with self.progress(
            message=f"Downloading evaluation data from {dir_name}",
            total=self.api.storage.get_directory_size(self.team_id, self.eval_dir),
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            self.api.storage.download_directory(
                self.team_id, self.eval_dir, self.local_dir, progress_cb=pbar.update
            )

    # def _load_projects(self):
    #     projects_dir = Path(self.local_dir, "projects")
    #     items_total = self.gt_images_ids
    #     if items_total is None:
    #         items_total = sum(self.gt_dataset_infos, key=lambda x: x.items_count)
    #     with self.progress(
    #         message=f"Downloading GT project {self.gt_project_info.name} and datasets",
    #         total=items_total,
    #     ) as pbar:
    #         download_project(
    #             self.api,
    #             self.gt_project_info.id,
    #             str(projects_dir),
    #             dataset_ids=self.gt_dataset_ids,
    #             progress_cb=pbar.update,
    #         )

    def _load_eval_data_archive(self, path: Path, pd_index_col: bool = False) -> Dict:
        """Load eval_data from archive"""
        with zipfile.ZipFile(path, mode="r") as zf:
            with zf.open("eval_data.json") as json_f:
                data = json.load(json_f)
            return self._process_value_from_archive(data, zf, pd_index_col=pd_index_col)

    def _process_value_from_archive(self, value, zf: zipfile.ZipFile, pd_index_col: bool = False):
        """Recursively process values from archive, handling nested dicts and lists."""
        if isinstance(value, str) and value.endswith(".npy"):
            with zf.open(value) as arr_f:
                return np.load(arr_f)
        elif isinstance(value, str) and value.endswith(".csv"):
            with zf.open(value) as df_f:
                if pd_index_col:
                    return pd.read_csv(df_f, sep="\t", index_col=0)
                return pd.read_csv(df_f, sep="\t")
        elif isinstance(value, dict):
            res = {}
            for k, v in value.items():
                k = int(k) if isinstance(k, str) and k.isdigit() else k
                k = float(k) if isinstance(k, str) and self._is_float(k) else k
                res[k] = self._process_value_from_archive(v, zf, pd_index_col=pd_index_col)
            return res
        elif isinstance(value, list):
            return [
                self._process_value_from_archive(item, zf, pd_index_col=pd_index_col)
                for item in value
            ]
        elif isinstance(value, str) and value.isdigit():
            return int(value)
        else:
            return value

    def _is_float(self, s: str) -> bool:
        if not s or not isinstance(s, str):
            return False
        try:
            float(s)
            return "." in s or "e" in s.lower()
        except (ValueError, AttributeError):
            return False

    def _read_eval_data(self):
        from pycocotools.coco import COCO  # pylint: disable=import-error

        gt_path = str(Path(self.local_dir, "evaluation", "cocoGt.json"))
        dt_path = str(Path(self.local_dir, "evaluation", "cocoDt.json"))
        coco_gt, coco_dt = COCO(gt_path), COCO(dt_path)
        self.coco_gt = coco_gt
        self.coco_dt = coco_dt
        eval_data_pickle_path = Path(self.local_dir, "evaluation", "eval_data.pkl")
        eval_data_archive_path = Path(self.local_dir, "evaluation", "eval_data.zip")
        if eval_data_pickle_path.exists():
            try:
                with open(eval_data_pickle_path, "rb") as f:
                    self.eval_data = pickle.load(f)
            except Exception as e:
                logger.warning(f"Failed to load eval_data.pkl: {e}", exc_info=True)
                self.eval_data = None
        if self.eval_data is None and eval_data_archive_path.exists():
            try:
                self.eval_data = self._load_eval_data_archive(eval_data_archive_path)
            except Exception as e:
                logger.warning(f"Failed to load eval_data from archive: {e}", exc_info=True)
                self.eval_data = None

        if self.eval_data is None:
            raise ValueError("Failed to load eval_data. Please contact support.")

        self.inference_info = load_json_file(
            Path(self.local_dir, "evaluation", "inference_info.json")
        )
        speedtest_info_path = Path(self.local_dir, "speedtest", "speedtest.json")
        if speedtest_info_path.exists():
            self.speedtest_info = load_json_file(
                Path(self.local_dir, "speedtest", "speedtest.json")
            )

        self.mp = MetricProvider(
            self.eval_data,
            self.coco_gt,
            self.coco_dt,
        )
        self.mp.calculate()

        self.df_score_profile = pd.DataFrame(
            self.mp.confidence_score_profile(), columns=["scores", "precision", "recall", "f1"]
        )

        # downsample
        if len(self.df_score_profile) > 5000:
            self.dfsp_down = self.df_score_profile.iloc[:: len(self.df_score_profile) // 1000]
        else:
            self.dfsp_down = self.df_score_profile

        self.f1_optimal_conf = self.mp.get_f1_optimal_conf()[0]
        if self.f1_optimal_conf is None:
            self.f1_optimal_conf = 0.01
            logger.warning("F1 optimal confidence cannot be calculated. Using 0.01 as default.")

        # Click data
        gt_id_mapper = IdMapper(self.coco_gt.dataset)
        dt_id_mapper = IdMapper(self.coco_dt.dataset)

        self.click_data = ClickData(self.mp.m, gt_id_mapper, dt_id_mapper)

    # def _update_comparison_data(
    #     self,
    #     gt_image_id: int,
    #     gt_image_info: ImageInfo = None,
    #     pred_image_info: ImageInfo = None,
    #     diff_image_info: ImageInfo = None,
    #     gt_annotation: Annotation = None,
    #     pred_annotation: Annotation = None,
    #     diff_annotation: Annotation = None,
    # ):
    #     comparison_data = self.comparison_data.get(gt_image_id, None)
    #     if comparison_data is None:
    #         self.comparison_data[gt_image_id] = ImageComparisonData(
    #             gt_image_info=gt_image_info,
    #             pred_image_info=pred_image_info,
    #             diff_image_info=diff_image_info,
    #             gt_annotation=gt_annotation,
    #             pred_annotation=pred_annotation,
    #             diff_annotation=diff_annotation,
    #         )
    #     else:
    #         for attr, value in {
    #             "gt_image_info": gt_image_info,
    #             "pred_image_info": pred_image_info,
    #             "diff_image_info": diff_image_info,
    #             "gt_annotation": gt_annotation,
    #             "pred_annotation": pred_annotation,
    #             "diff_annotation": diff_annotation,
    #         }.items():
    #             if value is not None:
    #                 setattr(comparison_data, attr, value)



from datetime import datetime
from pathlib import Path
from typing import Dict, Optional

import numpy as np

from supervisely import (
    Api,
    ObjClass,
    PointcloudAnnotation,
    PointcloudObject,
    Progress,
    ProjectMeta,
    TagMeta,
    TagValueType,
    TinyTimer,
    is_development,
    logger,
)
from supervisely.api.api import ApiField
from supervisely.convert.base_converter import AvailablePointcloudConverters
from supervisely.convert.pointcloud.lyft import lyft_helper
from supervisely.convert.pointcloud.pointcloud_converter import PointcloudConverter
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.io import fs
from supervisely.pointcloud_annotation.pointcloud_annotation import (
    PointcloudFigure,
    PointcloudObjectCollection,
    PointcloudTagCollection,
)
from supervisely.pointcloud_annotation.pointcloud_tag import PointcloudTag


class LyftConverter(PointcloudConverter):
    class Item(PointcloudConverter.Item):

        def __init__(
            self,
            item_path,
            ann_data: str = None,
            related_images: list = None,
            custom_data: dict = None,
            scene_name: str = None,
        ):
            super().__init__(item_path, ann_data, related_images, custom_data)
            self._type = "point_cloud"
            self._scene_name = scene_name

    def __init__(
        self,
        input_data: str,
        labeling_interface: str,
        upload_as_links: bool,
        remote_files_map: Optional[Dict[str, str]] = None,
    ):
        super().__init__(input_data, labeling_interface, upload_as_links, remote_files_map)
        self._is_pcd_episode = False
        self._lyft = None

    def __str__(self) -> str:
        return AvailablePointcloudConverters.LYFT

    @property
    def key_file_ext(self) -> str:
        return ".bin"

    def validate_format(self) -> bool:
        try:
            from lyft_dataset_sdk.lyftdataset import LyftDataset as Lyft
        except ImportError:
            logger.warn(
                'Install "lyft_dataset_sdk" python package to import datasets in LYFT format.'
            )
            return False

        def filter_fn(path):
            return all([(Path(path) / name).exists() for name in lyft_helper.FOLDER_NAMES])

        input_paths = [d for d in fs.dirs_filter(self._input_data, filter_fn)]
        if len(input_paths) == 0:
            return False
        input_path = input_paths[0]

        lidar_dir = input_path + "/lidar/"
        json_dir = input_path + "/data/"
        if lyft_helper.validate_ann_dir(json_dir) is False:
            return False

        bin_files = fs.list_files_recursively(
            lidar_dir, [self.key_file_ext], ignore_valid_extensions_case=True
        )

        if len(bin_files) == 0:
            return False

        # check if pointclouds have 5 columns (x, y, z, intensity, ring)
        pointcloud = np.fromfile(bin_files[0], dtype=np.float32)
        if pointcloud.shape[0] % 5 != 0:
            return False

        try:
            t = TinyTimer()
            lyft = Lyft(data_path=input_path, json_path=json_dir, verbose=False)
            self._lyft: Lyft = lyft
            logger.info(f"LyftDataset initialization took {t.get_sec():.2f} sec")
        except Exception as e:
            logger.info(f"Failed to initialize LyftDataset: {e}")
            return False

        t = TinyTimer()
        progress = Progress(f"Extracting annotations from available scenes...")
        # i = 0 # for debug
        for scene in lyft_helper.get_available_scenes(lyft):
            scene_name = scene["name"]
            sample_datas = lyft_helper.extract_data_from_scene(lyft, scene)
            if sample_datas is None:
                logger.warning(f"Failed to extract sample data from scene: {scene['name']}.")
                continue
            for sample_data in sample_datas:
                item_path = sample_data["lidar_path"]
                ann_data = sample_data["ann_data"]
                related_images = lyft_helper.get_related_images(ann_data)
                custom_data = sample_data.get("custom_data", {})
                item = self.Item(item_path, ann_data, related_images, custom_data, scene_name)
                self._items.append(item)
            progress.iter_done_report()
            # i += 1
            # if i == 2:
            #     break
        t = t.get_sec()
        logger.info(
            f"Lyft annotation extraction took {t:.2f} sec ({(t / self.items_count):.3f} sec per sample)"
        )

        return self.items_count > 0

    def to_supervisely(
        self,
        item: PointcloudConverter.Item,
        meta: ProjectMeta,
        renamed_classes: dict = {},
        renamed_tags: dict = {},
    ) -> PointcloudAnnotation:
        """
        Converts a point cloud item and its annotations to the supervisely formats.

        Args:
            item (PointcloudConverter.Item): The point cloud item to convert.
            meta (ProjectMeta): The project meta.

        Returns:
            PointcloudAnnotation: The converted point cloud annotation.
        """
        import open3d as o3d  # pylint: disable=import-error

        if getattr(item, "ann_data", None) is None:
            return PointcloudAnnotation()

        data = item.ann_data

        # * Convert annotation to json
        boxes = data["gt_boxes"]
        names = data["names"]

        objects = []
        for name, box in zip(names, boxes):
            center = [float(box[0]), float(box[1]), float(box[2])]
            size = [float(box[3]), float(box[5]), float(box[4])]
            ry = float(box[6])

            yaw = ry - np.pi
            yaw = yaw - np.floor(yaw / (2 * np.pi) + 0.5) * 2 * np.pi
            world_cam = None
            objects.append(o3d.ml.datasets.utils.BEVBox3D(center, size, yaw, name, -1.0, world_cam))
            objects[-1].yaw = ry

        geoms = [lyft_helper._convert_BEVBox3D_to_geometry(box) for box in objects]

        figures = []
        objs = []
        for l, geometry, token in zip(
            objects, geoms, data["instance_tokens"]
        ):  # by object in point cloud
            class_name = renamed_classes.get(l.label_class, l.label_class)
            tag_names = [
                self._lyft.get("attribute", attr_token).get("name", None)
                for attr_token in token["attribute_tokens"]
            ]
            tag_col = None
            if len(tag_names) > 0 and all([tag_name is not None for tag_name in tag_names]):
                tag_meta_names = [renamed_tags.get(name, name) for name in tag_names]
                tag_metas = [meta.get_tag_meta(tag_meta_name) for tag_meta_name in tag_meta_names]
                tag_col = PointcloudTagCollection([PointcloudTag(meta, None) for meta in tag_metas])
            pcobj = PointcloudObject(meta.get_obj_class(class_name), tag_col)
            figures.append(PointcloudFigure(pcobj, geometry))
            objs.append(pcobj)
        return PointcloudAnnotation(PointcloudObjectCollection(objs), figures)

    def upload_dataset(self, api: Api, dataset_id: int, batch_size: int = 1, log_progress=True):
        unique_names = {name for item in self._items for name in item.ann_data["names"]}
        tag_names = {tag["name"] for tag in self._lyft.attribute}
        self._meta = ProjectMeta(
            [ObjClass(name, Cuboid3d) for name in unique_names],
            [TagMeta(tag, TagValueType.NONE) for tag in tag_names],
        )
        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        scene_names = set([item._scene_name for item in self._items])
        dataset_info = api.dataset.get_info_by_id(dataset_id)
        scene_name_to_dataset = {}

        multiple_scenes = len(scene_names) > 1
        if multiple_scenes:
            logger.info(
                f"Found {len(scene_names)} scenes ({self.items_count} pointclouds) in the input data."
            )
            # * Create a nested dataset for each scene
            for name in scene_names:
                ds = api.dataset.create(
                    dataset_info.project_id,
                    name,
                    change_name_if_conflict=True,
                    parent_id=dataset_id,
                )
                scene_name_to_dataset[name] = ds
        else:
            scene_name_to_dataset[list(scene_names)[0]] = dataset_info

        if log_progress:
            progress, progress_cb = self.get_progress(self.items_count, "Converting pointclouds...")
        else:
            progress_cb = None

        for item in self._items:
            # * Get the current dataset for the scene
            current_dataset = scene_name_to_dataset.get(item._scene_name, None)
            if current_dataset is None:
                raise RuntimeError(f"Dataset not found for scene name: {item._scene_name}")
            current_dataset_id = current_dataset.id

            # * Convert timestamp to ISO format
            iso_time = datetime.utcfromtimestamp(item.ann_data["timestamp"] / 1e6).isoformat() + "Z"
            item.ann_data["timestamp"] = iso_time

            # * Convert pointcloud from ".bin" to ".pcd"
            pcd_path = str(Path(item.path).with_suffix(".pcd"))
            if fs.file_exists(pcd_path):
                logger.warning(f"Overwriting file with path: {pcd_path}")
            lyft_helper.convert_bin_to_pcd(item.path, pcd_path)

            # * Upload pointcloud
            pcd_name = fs.get_file_name(pcd_path)
            info = api.pointcloud.upload_path(current_dataset_id, pcd_name, pcd_path, {})
            pcd_id = info.id

            # * Convert annotation and upload
            ann = self.to_supervisely(item, meta, renamed_classes, renamed_tags)
            api.pointcloud.annotation.append(pcd_id, ann)

            # * Upload related images
            image_jsons = []
            camera_names = []
            for img_path, rimage_info in lyft_helper.generate_rimage_infos(
                item._related_images, item.ann_data
            ):
                img = api.pointcloud.upload_related_image(img_path)
                image_jsons.append(
                    {
                        ApiField.ENTITY_ID: pcd_id,
                        ApiField.NAME: rimage_info[ApiField.NAME],
                        ApiField.HASH: img,
                        ApiField.META: rimage_info[ApiField.META],
                    }
                )
                camera_names.append(rimage_info[ApiField.META]["deviceId"])
            if len(image_jsons) > 0:
                api.pointcloud.add_related_images(image_jsons, camera_names)

            # * Clean up
            fs.silent_remove(pcd_path)
            if log_progress:
                progress_cb(1)

        logger.info(f"Dataset ID:{current_dataset_id} has been successfully uploaded.")

        if log_progress:
            if is_development():
                progress.close()



from typing import List, Union
import supervisely as sly


class ClassMap:
    def __init__(self, obj_classes: Union[sly.ObjClassCollection, List[sly.ObjClass]]):
        self.obj_classes = obj_classes
        self.class2idx = {obj_class.name: idx for idx, obj_class in enumerate(self.obj_classes)}
        self.idx2class = {idx: obj_class.name for idx, obj_class in enumerate(self.obj_classes)}
        self.classes = [obj_class.name for obj_class in self.obj_classes]
        self.sly_ids = [obj_class.sly_id for obj_class in self.obj_classes]
    
    def __len__(self):
        return len(self.obj_classes)


from typing import Any, Dict, List

from supervisely.api.api import Api
from supervisely.app.widgets import Button, Card, Container, DeployModel, Text


class ModelSelector:
    title = "Model"
    description = "Connect to deployed model or deploy new model"
    lock_message = "Select previous step to unlock"

    def __init__(self, api: Api, team_id: int):
        # Init Step
        self.api = api
        self.team_id = team_id
        self.display_widgets: List[Any] = []
        # -------------------------------- #

        # Init Base Widgets
        self.validator_text = None
        self.button = None
        self.container = None
        self.card = None
        # -------------------------------- #

        # Model Selector
        self.model = DeployModel(api=self.api, team_id=self.team_id)
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.model])
        # ----------------------------------- #

        # Base Widgets
        self.validator_text = Text("")
        self.validator_text.hide()
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.validator_text])
        # ----------------------------------- #

        # Card Layout
        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
        )
        # ----------------------------------- #

    @property
    def widgets_to_disable(self) -> list:
        return [
            self.model,
            self.model.connect_button,
            self.model.deploy_button,
            self.model.stop_button,
            self.model.disconnect_button,
        ]

    def get_settings(self) -> Dict[str, Any]:
        return self.model.get_deploy_parameters()

    def load_from_json(self, data):
        self.model.load_from_json(data)

    def validate_step(self) -> bool:
        self.validator_text.hide()

        if self.model.model_api is None:
            self.validator_text.set(text="Please connect or deploy a model", status="error")
            self.validator_text.show()
            return False

        return True



# coding: utf-8
import mimetypes
from typing import Callable, List, Optional, Union

from requests_toolbelt import MultipartEncoder
from tqdm import tqdm

from supervisely import logger
from supervisely.api.module_api import ApiField, ModuleApiBase
from supervisely.collection.str_enum import StrEnum
from supervisely.io import env
from supervisely.io.fs import ensure_base_path, get_file_name_with_ext


class Provider(StrEnum):
    """Provider"""

    S3 = "s3"
    """S3"""
    GOOGLE = "google"
    """GOOGLE"""
    AZURE = "azure"
    """AZURE"""
    FS = "fs"
    """FS"""
    MINIO = "minio"
    """MINIO"""
    GCS = "gcs"
    """GCS"""

    @staticmethod
    def validate_path(path: str):
        if (
            not path.startswith(str(Provider.S3))
            and not path.startswith(str(Provider.GOOGLE))
            and not path.startswith(str(Provider.AZURE))
            and not path.startswith(str(Provider.FS))
            and not path.startswith(str(Provider.MINIO))
            and not path.startswith(str(Provider.GCS))
        ):
            prefix = path.split("://")[0]
            raise ValueError(
                f"Incorrect cloud provider '{prefix}' in path, learn more here: https://docs.supervisely.com/enterprise-edition/advanced-tuning/s3#links-plugin-cloud-providers-support"
            )


class RemoteStorageApi(ModuleApiBase):
    """RemoteStorageApi"""

    def _convert_json_info(self, info: dict):
        """_convert_json_info"""
        return info

    def get_file_info_by_path(
        self,
        path: str,
        team_id: int = None,
    ) -> Optional[dict]:
        """
        Get info about file for given remote path.

        :param path: Remote path to file.
        :type path: str
        :param team_id: Team ID (to get cloud storages connected to the team)
        :type team_id: int
        :returns: file  info in the given remote path
        :rtype: Optional[dict]

        """
        team_id = team_id or env.team_id(raise_not_found=False)
        Provider.validate_path(path)
        path = path.rstrip("/")

        json_body = {
            ApiField.PATH: path,
            ApiField.RECURSIVE: False,
            ApiField.FILES: True,
            ApiField.FOLDERS: False,
            ApiField.LIMIT: 1,
            "startAfter": "",
        }
        if team_id is not None:
            json_body[ApiField.GROUP_ID] = team_id

        resp = self._api.get("remote-storage.list", json_body)
        if resp is None:
            return None

        return resp.json()[0]

    def list(
        self,
        path: str,
        recursive: bool = True,
        files: bool = True,
        folders: bool = True,
        limit: int = 10000,
        start_after: str = "",
        team_id: int = None,
    ) -> list:
        """
        List files and directories for given remote path.

        :param path: Remote path with items that you want to list.
        :type path: str
        :param recursive: List remote path revursively.
        :type recursive: bool
        :param files: List files in the given path.
        :type files: bool
        :param folders: List folders in the given path.
        :type folders: bool
        :param limit: Limit of files to list. 10000 is the maximum limit.
        :type limit: int
        :param start_after: Start listing path after given file name.
        :type start_after: str
        :param team_id: Team ID (to get cloud storages connected to the team)
        :type team_id: int
        :returns: List of files in the given remote path
        :rtype: list

        """
        team_id = team_id or env.team_id(raise_not_found=False)

        Provider.validate_path(path)
        path = path.rstrip("/") + "/"

        json_body = {
            ApiField.PATH: path,
            ApiField.RECURSIVE: recursive,
            ApiField.FILES: files,
            ApiField.FOLDERS: folders,
            ApiField.LIMIT: limit,
            "startAfter": start_after,
        }
        if team_id is not None:
            json_body[ApiField.GROUP_ID] = team_id

        resp = self._api.get("remote-storage.list", json_body)
        if resp is None:
            return []
        return resp.json()

    def download_path(
        self,
        remote_path: str,
        save_path: str,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        team_id: int = None,
    ):
        """
        Downloads item from given remote path to given local path.

        :param remote_path: Remote path to item that you want to download.
        :type remote_path: str
        :param save_path: Local save path.
        :type save_path: str
        :param progress_cb: Progress function to download.
        :type progress_cb: tqdm or callable, optional
        :param team_id: Team ID (to get cloud storages connected to the team)
        :type team_id: int


        .. code-block:: python

            provider = "s3" # can be one of ["s3", "google", "azure"]
            bucket = "bucket-test-export"
            path_in_bucket = "/demo/image.jpg"
            remote_path = api.remote_storage.get_remote_path(provider, bucket, path_in_bucket)
            # or alternatively use this:
            # remote_path = f"{provider}://{bucket}{path_in_bucket}"
            api.remote_storage.download_path(local_path="images/my-cats.jpg", remote_path=remote_path)
        """
        team_id = team_id or env.team_id(raise_not_found=False)
        Provider.validate_path(remote_path)
        ensure_base_path(save_path)

        json_body = {ApiField.LINK: remote_path}
        if team_id is not None:
            json_body[ApiField.GROUP_ID] = team_id

        response = self._api.post("remote-storage.download", json_body, stream=True)
        # if "Content-Length" in response.headers:
        #     length = int(response.headers['Content-Length'])
        with open(save_path, "wb") as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)
                if progress_cb is not None:
                    progress_cb(len(chunk))

    def upload_path(self, local_path: str, remote_path: str, team_id: int = None):
        """
        Uploads item from given local path to given remote path.

        :param local_path: Local path to item that you want to upload.
        :type local_path: str
        :param remote_path: Remote destination path.
        :type remote_path: str
        :param team_id: Team ID (to get cloud storages connected to the team)
        :type team_id: int
        :Usage example:

        .. code-block:: python

            provider = "s3" # can be one of ["s3", "google", "azure"]
            bucket = "bucket-test-export"
            path_in_bucket = "/demo/image.jpg"
            remote_path = api.remote_storage.get_remote_path(provider, bucket, path_in_bucket)
            team_id = 123
            # or alternatively use this:
            # remote_path = f"{provider}://{bucket}{path_in_bucket}"
            api.remote_storage.upload_path("images/my-cats.jpg", remote_path, team_id)
        """
        Provider.validate_path(remote_path)
        return self._upload_paths_batch([local_path], [remote_path], team_id)

    def _upload_paths_batch(self, local_paths, remote_paths, team_id: int = None):
        """_upload_paths_batch"""
        team_id = team_id or env.team_id(raise_not_found=False)

        if len(local_paths) != len(remote_paths):
            raise ValueError("Inconsistency in paths, len(local_paths) != len(remote_paths)")
        if len(local_paths) == 0:
            return {}

        def path_to_bytes_stream(path):
            return open(path, "rb")

        content = []
        for idx, (src, dst) in enumerate(zip(local_paths, remote_paths)):
            content.append((ApiField.PATH, dst))
            name = get_file_name_with_ext(dst)
            content.append(
                (
                    "file",
                    (
                        name,
                        path_to_bytes_stream(src),
                        mimetypes.MimeTypes().guess_type(src)[0],
                    ),
                )
            )
        encoder = MultipartEncoder(fields=content)
        url = f"remote-storage.upload"
        if team_id is not None:
            url += f"?teamId={team_id}"
        resp = self._api.post(url, encoder)
        return resp.json()

    def get_remote_path(self, provider: str, bucket: str, path_in_bucket: str) -> str:
        """
        Returns remote path.

        :param provider: Can be one of "s3", "google", "azure".
        :type provider: str
        :param bucket: Name of the bucket container.
        :type bucket: str
        :param path_in_bucket: Path to item in bucket.
        :type path_in_bucket: str
        :Usage example:

        .. code-block:: python

            provider = "s3"
            bucket = "bucket-test-export"
            path_in_bucket = "/demo/image.jpg"
            remote_path = api.remote_storage.get_remote_path(provider, bucket, path_in_bucket)
            # Output: s3://bucket-test-export/demo/image.jpg
        """
        res_path = f"{provider}://{bucket}/{path_in_bucket.lstrip('/')}"
        Provider.validate_path(res_path)
        return res_path

    def get_list_available_providers(
        self,
        team_id: int = None,
    ) -> List[dict]:
        """
        Get the list of available providers for the instance.

        :param team_id: Team ID (to get cloud storages connected to the team)
        :type team_id: int
        :return: List of available providers
        :rtype: List[dict]
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
               load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            # Pass values into the API constructor (optional, not recommended)
            # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

            team_id = 123
            available_providers = api.remote_storage.get_list_available_providers(team_id)

            # Output example

            #    [
            #        {
            #            "id": "minio",
            #            "name": "Amazon S3",
            #            "defaultProtocol": "s3:",
            #            "protocols": [
            #                "s3:",
            #                "minio:"
            #            ],
            #            "buckets": [
            #                "bucket-test",
            #                "remote-img"
            #            ]
            #        }
            #    ]

        """
        team_id = team_id or env.team_id(raise_not_found=False)

        json_body = {}
        if team_id is not None:
            json_body[ApiField.GROUP_ID] = team_id

        resp = self._api.get("remote-storage.available_providers", json_body)
        return resp.json()

    def get_list_supported_providers(
        self,
        team_id: int = None,
    ) -> List[dict]:
        """
        Get the list of supported providers for the instance.

        :param team_id: Team ID (to get cloud storages connected to the team)
        :type team_id: int
        :return: List of supported providers
        :rtype: List[dict]
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
               load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            # Pass values into the API constructor (optional, not recommended)
            # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

            team_id = 123
            supported_providers = api.remote_storage.get_list_supported_providers(team_id)

            # Output example

            #    [
            #        {
            #            "id": "google",
            #            "name": "Google Cloud Storage",
            #            "defaultProtocol": "google:",
            #            "protocols": [
            #                "google:",
            #                "gcs:"
            #            ]
            #        }
            #    ]

        """
        team_id = team_id or env.team_id(raise_not_found=False)

        json_body = {}
        if team_id is not None:
            json_body[ApiField.GROUP_ID] = team_id

        resp = self._api.get("remote-storage.supported_providers", json_body)
        return resp.json()

    def is_path_exist(
        self,
        path: str,
        team_id: int = None,
    ) -> bool:
        """
        Check if the file path exists.

        :param path: URL of the file in the bucket storage
        :type path: str
        :param team_id: Team ID (to get cloud storages connected to the team)
        :type team_id: int
        :return: True if the file exists, False otherwise
        :rtype: bool
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
               load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            # Pass values into the API constructor (optional, not recommended)
            # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

            path = "s3://bucket/lemons/ds1/img/IMG_444.jpeg"
            is_exist = api.remote_storage.is_path_exist(path)

        """
        team_id = team_id or env.team_id(raise_not_found=False)
        Provider.validate_path(path)

        json_body = {ApiField.PATH: path}
        if team_id is not None:
            json_body[ApiField.GROUP_ID] = team_id

        resp = self._api.get("remote-storage.exists", json_body)
        if resp is None:
            return False

        resp = resp.json()

        if resp.get("exists"):
            return True
        else:
            return False

    def get_path_stats(
        self,
        path: str,
        team_id: int = None,
    ) -> Optional[dict]:
        """
        Get information about file size and the date of its last modification in bucket storage.

        :param path: URL of the file in the bucket storage
        :type path: str
        :param team_id: Team ID (to get cloud storages connected to the team)
        :type team_id: int
        :return: File 'size' in bytes and 'lastModified' date if file exists, otherwise None
        :rtype: Optional[dict]
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
               load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            # Pass values into the API constructor (optional, not recommended)
            # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

            path = "s3://bucket/lemons/ds1/img/IMG_444.jpeg"
            team_id = 123
            stats = api.remote_storage.get_path_stats(path, team_id)

            # Output example

            #   {
            #       "size": 155790,
            #       "lastModified": "2023-01-26T09:20:27.000Z"
            #   }

        """
        team_id = team_id or env.team_id(raise_not_found=False)

        json_body = {ApiField.PATH: path}
        if team_id is not None:
            json_body[ApiField.GROUP_ID] = team_id

        if self.is_path_exist(path, team_id):
            resp = self._api.get("remote-storage.stat", json_body)
            return resp.json()
        else:
            path_folers = path.split("/")[3:]
            file_path = ""
            for part in path_folers:
                file_path += part + "/"
            file_path = file_path.rstrip("/")
            logger.warning(f"The file doesn't exist! Check the path: {file_path}")
            return None

    @staticmethod
    def is_bucket_url(url: str) -> bool:
        """
        Check if the URL is a bucket URL.

        :param url: URL
        :type url: str
        :return: True if URL is a bucket URL, False otherwise
        :rtype: bool

        :Usage example:

         .. code-block:: python

            from supervisely.api.remote_storage_api import RemoteStorageApi

            url = "s3://bucket/lemons/ds1/img/IMG_444.jpeg"
            RemoteStorageApi.is_bucket_url(url)

        """
        provider_protocols = [
            Provider.S3.value,
            Provider.MINIO.value,
            Provider.GOOGLE.value,
            Provider.GCS.value,
            Provider.AZURE.value,
            Provider.FS.value,
        ]
        return any(url.startswith(protocol + "://") for protocol in provider_protocols)



# coding: utf-8

from typing import Any, Dict, List, Optional, Union

from supervisely._utils import batched
from supervisely.api.module_api import ApiField, ModuleApi
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_settings import LabelingInterface
from supervisely.task.progress import tqdm_sly
from supervisely.video_annotation.key_id_map import KeyIdMap


class TagApi(ModuleApi):
    """"""

    _entity_id_field = None
    """"""
    _method_bulk_add = None
    """"""

    @staticmethod
    def info_sequence():
        """
        NamedTuple TagInfo information about Tag.

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            info_sequence = api.video.tag.info_sequence()
        """

        return [
            ApiField.ID,
            ApiField.PROJECT_ID,
            ApiField.NAME,
            ApiField.SETTINGS,
            ApiField.COLOR,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of NamedTuple for class.

        :return: NamedTuple name.
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            tuple_name = api.video.tag.info_tuple_name()
            print(tuple_name) # TagInfo
        """

        return "TagInfo"

    def get_list(self, project_id: int, filters=None):
        """
        Get list of tags for a given project ID.

        :param project_id: :class:`Dataset<supervisely.project.project.Project>` ID in Supervisely.
        :type project_id: int
        :param filters: List of parameters to sort output tags. See: https://api.docs.supervisely.com/#tag/Advanced/paths/~1tags.list/get
        :type filters: List[Dict[str, str]], optional
        :return: List of the tags from the project with given id.
        :rtype: list
        """

        return self.get_list_all_pages(
            "tags.list", {ApiField.PROJECT_ID: project_id, "filter": filters or []}
        )

    def get_name_to_id_map(self, project_id: int):
        """
        Get dictionary with mapping tag name to tag ID for a given project ID.

        :param project_id: :class:`Dataset<supervisely.project.project.Project>` ID in Supervisely.
        :type project_id: int
        :return: Dictionary with mapping tag name to tag id for a given project ID.
        :rtype: dict
        """

        tags_info = self.get_list(project_id)
        return {tag_info.name: tag_info.id for tag_info in tags_info}

    def _tags_to_json(self, tags: KeyIndexedCollection, tag_name_id_map=None, project_id=None):
        """"""
        if tag_name_id_map is None and project_id is None:
            raise RuntimeError("Impossible to get ids for project tags")
        if tag_name_id_map is None:
            tag_name_id_map = self.get_name_to_id_map(project_id)
        tags_json = []
        tags_keys = []
        for tag in tags:
            tag_json = tag.to_json()
            tag_json[ApiField.TAG_ID] = tag_name_id_map[tag.name]
            tags_json.append(tag_json)
            tags_keys.append(tag.key())
        return tags_json, tags_keys

    def append_to_entity(
        self,
        entity_id: int,
        project_id: int,
        tags: KeyIndexedCollection,
        key_id_map: KeyIdMap = None,
    ):
        """
        Add tags to entity in project with given ID.

        :param entity_id: ID of the entity in Supervisely to add a tag to
        :type entity_id: int
        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param tags: Collection of tags
        :type tags: KeyIndexedCollection
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: List of tags IDs
        :rtype: list
        """

        if len(tags) == 0:
            return []
        tags_json, tags_keys = self._tags_to_json(tags, project_id=project_id)
        ids = self._append_json(entity_id, tags_json)
        KeyIdMap.add_tags_to(key_id_map, tags_keys, ids)
        return ids

    def _append_json(self, entity_id, tags_json):
        """"""
        if self._method_bulk_add is None:
            raise RuntimeError("self._method_bulk_add is not defined in child class")
        if self._entity_id_field is None:
            raise RuntimeError("self._entity_id_field is not defined in child class")

        if len(tags_json) == 0:
            return []
        response = self._api.post(
            self._method_bulk_add, {self._entity_id_field: entity_id, ApiField.TAGS: tags_json}
        )
        ids = [obj[ApiField.ID] for obj in response.json()]
        return ids

    def append_to_objects(
        self,
        entity_id: int,
        project_id: int,
        objects: KeyIndexedCollection,
        key_id_map: KeyIdMap,
        is_video_multi_view: bool = False,
    ):
        """
        Add Tags to Annotation Objects for a specific entity (image etc.).

        :param entity_id: ID of the entity in Supervisely to add a tag to its objects
        :type entity_id: int
        :param project_id: Project ID in Supervisely. Uses to get tag name to tag ID mapping.
        :type project_id: int
        :param objects: Collection of Annotation Objects.
        :type objects: KeyIndexedCollection
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :param is_video_multi_view: If True, indicates that the entity is a multi-view video.
        :type is_video_multi_view: bool
        :return: List of tags IDs
        :rtype: list
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_id = 19373170
            img_info = api.
        """

        tag_name_id_map = self.get_name_to_id_map(project_id)

        tags_to_add = []
        tags_keys = []
        for object in objects:
            obj_id = key_id_map.get_object_id(object.key())
            if obj_id is None:
                raise RuntimeError(
                    "Can not add tags to object: OBJECT_ID not found for key {}".format(
                        object.key()
                    )
                )
            tags_json, cur_tags_keys = self._tags_to_json(
                object.tags, tag_name_id_map=tag_name_id_map
            )
            for tag in tags_json:
                tag[ApiField.OBJECT_ID] = obj_id
                tags_to_add.append(tag)
            tags_keys.extend(cur_tags_keys)

        if len(tags_keys) != len(tags_to_add):
            raise RuntimeError("SDK error: len(tags_keys) != len(tags_to_add)")
        if len(tags_keys) == 0:
            return
        ids = self.append_to_objects_json(entity_id, tags_to_add, project_id, is_video_multi_view)
        KeyIdMap.add_tags_to(key_id_map, tags_keys, ids)
        return ids

    def append_to_objects_json(
        self,
        entity_id: int,
        tags_json: List[Dict],
        project_id: Optional[int] = None,
        is_video_multi_view: bool = False,
    ) -> List[int]:
        """
        Add Tags to Annotation Objects for specific entity (image etc.).

        :param entity_id: ID of the entity in Supervisely to add a tag to its objects
        :type entity_id: int
        :param tags_json: Collection of tags in JSON format
        :type tags_json: dict
        :param project_id: Project ID in Supervisely. Uses to get tag name to tag ID mapping.
                           Not required if `multi_view` is True.
        :type project_id: int, optional
        :param is_video_multi_view: If True, indicates that the entity is a multi-view video.
        :type is_video_multi_view: bool
        :return: List of tags IDs
        :rtype: list

        :Usage example:

             .. code-block:: python

                import supervisely as sly

                api = sly.Api(server_address, token)

                tags_list = [
                    {
                        "tagId": 25926,
                        "objectId": 652959,
                        "value": None
                    },
                    {
                        "tagId": 25927,
                        "objectId": 652959,
                        "value": "v1"
                    },
                    {
                        "tagId": 25927,
                        "objectId": 652958,
                        "value": "v2"
                    }
                ]
                response = api.video.tag.append_to_objects_json(12345, tags_list)

                print(response)
                # Output:
                #    [
                #        80421101,
                #        80421102,
                #        80421103
                #    ]
        """
        project_meta = self._api.optimization_context.get("project_meta")

        if isinstance(project_meta, ProjectMeta):
            if project_meta.labeling_interface == LabelingInterface.MULTIVIEW:
                is_video_multi_view = True

        if len(tags_json) == 0:
            return []
        if project_id is not None and not is_video_multi_view:
            json_data = {ApiField.PROJECT_ID: project_id, ApiField.TAGS: tags_json}
        else:
            json_data = {ApiField.ENTITY_ID: entity_id, ApiField.TAGS: tags_json}
        response = self._api.post("annotation-objects.tags.bulk.add", json_data)
        ids = [obj[ApiField.ID] for obj in response.json()]
        return ids

    def add_to_objects(
        self,
        project_id: int,
        tags_list: List[dict],
        batch_size: int = 100,
        log_progress: bool = False,
        progress: Optional[tqdm_sly] = None,
        is_video_multi_view: bool = False,
        entity_id: Optional[int] = None,
    ) -> List[Dict[str, Union[str, int, None]]]:
        """
        For images project:
            Add Tags to existing Annotation Figures (labels).
            The `tags_list` example:
                [{"tagId": 12345, "figureId": 54321, "value": "tag_value"}, ...].
        For video, pointcloud, volume and pointcloud episodes projects:
            Add Tags to existing Annotation Objects.
            The `frameRange` field is optional and is supported only for video and pointcloud episodes projects.
            The `tags_list`` example:
                [{"tagId": 12345, "objectId": 54321, "value": "tag_value"}, ...].
            or with frameRange:
                [{"tagId": 12345, "objectId": 54321, "value": "tag_value", "frameRange": [1, 10]}, ...].

        All objects must belong to entities of the same project.

        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param tags_list: List of tag object infos as dictionaries.
        :type tags_list: List[dict]
        :param batch_size: Number of tags to add in one request.
        :type batch_size: int
        :param log_progress: If True, will display a progress bar.
        :type log_progress: bool
        :param progress: Progress bar object to display progress.
        :type progress: Optional[tqdm_sly]
        :param is_video_multi_view: If True, indicates that the entity is a multi-view video.
        :type is_video_multi_view: bool
        :param entity_id: ID of the entity in Supervisely to add a tag to its objects.
                          Required if `is_video_multi_view` is True.
        :type entity_id: Optional[int]
        :return: List of tags infos as dictionaries.
        :rtype: List[Dict[str, Union[str, int, None]]]

        Usage example:
        .. code-block:: python

            import supervisely as sly

            api = sly.Api(server_address, token)

            tag_list = [
                {
                    "tagId": 25926,
                    "figureId": 652959,
                    "value": None # optional for tag with type 'None'
                    "frameRange": [1, 10] # optional (supported only for video and pointcloud episodes projects)
                },
                {
                    "tagId": 25927,
                    "figureId": 652959,
                    "value": "v1"
                },
                {
                    "tagId": 25927,
                    "figureId": 652958,
                    "value": "v2",
                }
            ]
            response = api.image.tag.add_to_figures(12345, tag_list)

            print(response)
            # Output:
            #    [
            #        {
            #            "id": 80421101,
            #            "tagId": 25926,
            #            "figureId": 652959,
            #            "value": None
            #        },
            #        {
            #            "id": 80421102,
            #            "tagId": 25927,
            #            "figureId": 652959,
            #            "value": "v1"
            #        },
            #        {
            #            "id": 80421103,
            #            "tagId": 25927,
            #            "figureId": 652958,
            #            "value": "v2"
            #        }
            #    ]
        """

        if progress is not None:
            log_progress = False

        project_meta = self._api.optimization_context.get("project_meta")

        if isinstance(project_meta, ProjectMeta):
            if project_meta.labeling_interface == LabelingInterface.MULTIVIEW:
                is_video_multi_view = True

        result = []

        if len(tags_list) == 0:
            return result
        if log_progress:
            progress = tqdm_sly(
                desc="Adding tags to figures",
                total=len(tags_list),
            )
        for batch in batched(tags_list, batch_size):
            if is_video_multi_view:
                if entity_id is None:
                    raise ValueError("entity_id must be provided when is_video_multi_view is True")
                data = {ApiField.ENTITY_ID: entity_id, ApiField.TAGS: batch}
            else:
                data = {ApiField.PROJECT_ID: project_id, ApiField.TAGS: batch}
            if type(self) is TagApi:
                response = self._api.post("figures.tags.bulk.add", data)
            else:
                response = self._api.post("annotation-objects.tags.bulk.add", data)
            result.extend(response.json())
            if progress is not None:
                progress.update(len(batch))
        return result

    def add_to_entities_json(
        self,
        project_id: int,
        tags_list: List[Dict[str, Union[str, int, None]]],
        batch_size: int = 100,
        log_progress: bool = False,
    ) -> List[int]:
        """
        Bulk add tags to entities (images, videos, pointclouds, volumes) in a project.
        Not supported for pointcloud episodes projects.
        All entities must belong to the same project.
        The `frameRange` field in a tag object within the tags list is optional and is supported only for video projects.

        The `tags_list` example:
            [{"tagId": 12345, "entityId": 54321, "value": "tag_value"}, ...].
        or with frameRange:
            [{"tagId": 12345, "entityId": 54321, "value": "tag_value", "frameRange": [1, 10]}, ...].

        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param tags_list: List of tag object infos as dictionaries
                          (e.g. {"tagId": 12345, "entityId": 54321, "value": "tag_value"}).
        :param batch_size: Number of tags to add in one request.
        :type batch_size: int
        :param log_progress: If True, will display a progress bar.
        :type log_progress: bool
        :return: List of tags IDs.
        :rtype: List[int]

        Usage example:
        .. code-block:: python

            import supervisely as sly

            api = sly.Api(server_address, token)

            tag_list = [
                {
                    "tagId": 25926,
                    "entityId": 652959,
                    "value": None # optional for tag with type 'None'
                    "frameRange": [1, 10] # optional (supported only for video projects)
                },
                {
                    "tagId": 25927,
                    "entityId": 652959,
                    "value": "v1"
                },
                {
                    "tagId": 25927,
                    "entityId": 652958,
                    "value": "v2"
                }
            ]
            api.image.tag.add_to_entities_json(project_id=12345, tag_list=tag_list)
        """

        result = []

        if len(tags_list) == 0:
            return result

        if log_progress:
            ds_progress = tqdm_sly(desc="Adding tags to entities", total=len(tags_list))

        for batch in batched(tags_list, batch_size):
            data = {ApiField.PROJECT_ID: project_id, ApiField.TAGS: batch}
            response = self._api.post("tags.entities.bulk.add", data)
            result.extend([obj[ApiField.ID] for obj in response.json()])
            if log_progress:
                ds_progress.update(len(batch))

        return result

    def add_tags_collection_to_objects(
        self,
        project_id: int,
        tags_map: Dict[int, Any],
        batch_size: int = 100,
        log_progress: bool = False,
        is_video_multi_view: bool = False,
        entity_id: Optional[int] = None,
    ) -> List[Dict[str, Union[str, int, None]]]:
        """
        For images project:
            Add Tags to existing Annotation Figures (labels).
            The `tags_map` example: {figure_id_1: TagCollection, ...}.
        For video, pointcloud, volume and pointcloud episodes projects:
            Add Tags to existing Annotation Objects.
            The `frameRange` field is optional and is supported only for video and pointcloud episodes projects.
            The `tags_map` example: {object_id_1: TagCollection, ...}.

        All objects must belong to entities of the same project.

        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param tags_map: Dictionary with mapping figure/object ID to tags collection.
        :type tags_map: Dict[int, Any]
        :param batch_size: Number of tags to add in one request.
        :type batch_size: int
        :param log_progress: If True, will display a progress bar.
        :type log_progress: bool
        :param is_video_multi_view: If True, indicates that the entity is a multi-view video.
        :type is_video_multi_view: bool
        :param entity_id: ID of the entity in Supervisely to add a tag to its objects.
                          Required if `is_video_multi_view` is True.
        :type entity_id: Optional[int]
        :return: List of tags infos as dictionaries.
        :rtype: List[Dict[str, Union[str, int, None]]]

        Usage example:
        .. code-block:: python

            import supervisely as sly

            api = sly.Api(server_address, token)

            project_id = 12345

            tag_meta = sly.TagMeta("tag_name", sly.TagValueType.ANY_STRING)
            meta = sly.ProjectMeta(tag_metas=[tag_meta])
            meta = sly.ProjectMeta.from_json(api.project.update_meta(project_id, meta))
            tag_meta = meta.get_tag_meta("tag_name")

            # for images project:
            tag_map = {
                652959: sly.TagCollection([sly.Tag(tag_meta, value="v1"), sly.Tag(tag_meta, value="v2"), ...]),
                652958: sly.TagCollection([sly.Tag(tag_meta, value="v3"), sly.Tag(tag_meta, value="v4"), ...]),
                ...
            }
            api.image.tag.add_tags_to_objects(project_id, tag_map)

            # for videos projects (frameRange is optional):
            tag_map = {
                652959: sly.VideoTagCollection([sly.VideoTag(tag_meta, value="v1", frameRange=[1, 10]), ...]),
                652958: sly.VideoTagCollection([sly.VideoTag(tag_meta, value="v2", frameRange=[4, 12]), ...]),
                ...
            }
            api.video.tag.add_to_objects_json_batch(project_id, tag_map)
        """

        OBJ_ID_FIELD = ApiField.FIGURE_ID if type(self) is TagApi else ApiField.OBJECT_ID
        tag_name_id_map = self.get_name_to_id_map(project_id)

        data = []
        for obj_id, tags in tags_map.items():
            for tag in tags:
                tag_id = tag_name_id_map.get(tag.name)
                if tag_id is None:
                    raise ValueError(f"Tag {tag.name} not found in project {project_id}")

                tag_json = tag.to_json()
                tag_json[ApiField.TAG_ID] = tag_id
                tag_json[OBJ_ID_FIELD] = obj_id
                data.append(tag_json)

        return self.add_to_objects(
            project_id,
            data,
            batch_size,
            log_progress,
            is_video_multi_view=is_video_multi_view,
            entity_id=entity_id,
        )



import os
import shutil
from pathlib import Path
from typing import Callable, Dict, List, Optional, OrderedDict, Tuple, Union

import numpy as np
from PIL import Image
from tqdm import tqdm

from supervisely._utils import generate_free_name
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.tag import Tag, TagValueType
from supervisely.annotation.tag_meta import TagApplicableTo, TagMeta
from supervisely.convert.image.image_helper import validate_image_bounds
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging.color import generate_rgb
from supervisely.imaging.image import read
from supervisely.io.fs import file_exists, get_file_ext, get_file_name
from supervisely.io.json import load_json_file
from supervisely.project.project import Dataset, OpenMode, Project
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger
from supervisely.task.progress import tqdm_sly

MASKS_EXTENSION = ".png"

# Export
SUPPORTED_GEOMETRY_TYPES = {Bitmap, Polygon, Rectangle}
VALID_IMG_EXT = {".jpe", ".jpeg", ".jpg"}
TRAIN_TAG_NAME = "train"
VAL_TAG_NAME = "val"
TRAINVAL_TAG_NAME = "trainval"
DEFAULT_OBJECT_FIELDS = {"name", "class", "bndbox"}
DEFAULT_SUBCLASSES = {"pose", "truncated", "difficult", "occluded", "obstacle", "out-of-scope"}


default_classes_colors = {
    "neutral": (224, 224, 192),
    "aeroplane": (128, 0, 0),
    "bicycle": (0, 128, 0),
    "bird": (128, 128, 0),
    "boat": (0, 0, 128),
    "bottle": (128, 0, 128),
    "bus": (0, 128, 128),
    "car": (128, 128, 128),
    "cat": (64, 0, 0),
    "chair": (192, 0, 0),
    "cow": (64, 128, 0),
    "diningtable": (192, 128, 0),
    "dog": (64, 0, 128),
    "horse": (192, 0, 128),
    "motorbike": (64, 128, 128),
    "person": (192, 128, 128),
    "pottedplant": (0, 64, 0),
    "sheep": (128, 64, 0),
    "sofa": (0, 192, 0),
    "train": (128, 192, 0),
    "tvmonitor": (0, 64, 128),
}


# returns mapping: (r, g, b) color -> some (row, col) for each unique color except black
def get_col2coord(img: np.ndarray) -> dict:
    img = img.astype(np.int32)
    h, w = img.shape[:2]
    colhash = img[:, :, 0] * 256 * 256 + img[:, :, 1] * 256 + img[:, :, 2]
    unq, unq_inv, unq_cnt = np.unique(colhash, return_inverse=True, return_counts=True)
    indxs = np.split(np.argsort(unq_inv), np.cumsum(unq_cnt[:-1]))
    col2indx = {unq[i]: indxs[i][0] for i in range(len(unq))}
    return {
        (col // (256**2), (col // 256) % 256, col % 256): (indx // w, indx % w)
        for col, indx in col2indx.items()
        if col != 0
    }


def read_colors(colors_file: str) -> Tuple[ObjClassCollection, dict]:
    if os.path.isfile(colors_file):
        try:
            logger.info("Will try to read segmentation colors from provided file.")
            with open(colors_file, "r") as file:
                cls2col = {}
                for line in file:
                    parts = line.strip().split()
                    if len(parts) < 4:
                        raise ValueError("Invalid format of colors file.")
                    class_name = " ".join(parts[:-3])
                    colors = tuple(map(int, parts[-3:]))
                    cls2col[class_name] = colors
        except Exception as e:
            logger.warning(
                "Failed to read segmentation colors from provided file. "
                "Will use default PascalVOC color mapping."
            )
            cls2col = default_classes_colors
    else:
        logger.info("Will use default PascalVOC color mapping.")
        cls2col = default_classes_colors

    obj_classes_list = [
        ObjClass(name=class_name, geometry_type=Bitmap, color=color)
        for class_name, color in cls2col.items()
    ]

    logger.info(
        f"Determined {len(cls2col)} class(es).",
        extra={"classes": list(cls2col.keys())},
    )

    obj_classes = ObjClassCollection(obj_classes_list)
    color2class_name = {v: k for k, v in cls2col.items()}
    return obj_classes, color2class_name


def get_ann(
    item,
    color2class_name: dict,
    meta: ProjectMeta,
    bbox_classes_map: dict,
    renamed_classes=None,
    renamed_tags=None,
) -> Annotation:
    segm_path, inst_path = item.segm_path, item.inst_path
    height, width = item.shape
    img_rect = Rectangle.from_size(item.shape)
    ann = Annotation(img_size=(height, width))

    if item.ann_data is not None:
        bbox_labels = xml_to_sly_labels(
            item.ann_data, meta, bbox_classes_map, img_rect, renamed_classes, renamed_tags
        )
        ann = ann.add_labels(bbox_labels)

    if segm_path is None:
        return ann

    segmentation_img = read(segm_path)

    if inst_path is not None:
        instance_img = read(inst_path)
        colored_img = instance_img
        instance_img16 = instance_img.astype(np.uint16)
        col2coord = get_col2coord(instance_img16)
        curr_col2cls = []
        for col, coord in col2coord.items():
            cls_name = color2class_name.get(tuple(segmentation_img[coord]))
            if cls_name is not None:
                if renamed_classes is not None and cls_name in renamed_classes:
                    cls_name = renamed_classes[cls_name]
            curr_col2cls.append((col, cls_name))
        curr_col2cls = {
            k: v for k, v in curr_col2cls if v is not None
        }  # _instance_ color -> class name
    else:
        colored_img = segmentation_img
        segmentation_img = segmentation_img.astype(np.uint16)
        colors = list(get_col2coord(segmentation_img).keys())
        curr_col2cls = {}
        for color in colors:
            cls_name = color2class_name.get(color)
            if cls_name is not None:
                if renamed_classes is not None and cls_name in renamed_classes:
                    cls_name = renamed_classes[cls_name]
            curr_col2cls[color] = cls_name

    labels = []
    for color, class_name in curr_col2cls.items():
        mask = np.all(colored_img == color, axis=2)  # exact match (3-channel img & rgb color)
        bitmap = Bitmap(data=mask)
        obj_class = ObjClass(name=class_name, geometry_type=Bitmap)
        labels.append(Label(bitmap, obj_class))
        #  clear used pixels in mask to check missing colors, see below
        colored_img[mask] = (0, 0, 0)

    labels = validate_image_bounds(labels, img_rect)
    ann = ann.add_labels(labels)

    if np.sum(colored_img) > 0:
        logger.warning(
            f"Not all objects or classes are captured from source segmentation: {item.name}"
        )

    return ann


def xml_to_sly_labels(
    xml_path: str,
    meta: ProjectMeta,
    bbox_classes_map: dict,
    img_rect: Rectangle,
    renamed_classes=None,
    renamed_tags=None,
) -> List[Label]:
    import xml.etree.ElementTree as ET

    labels = []
    with open(xml_path, "r") as in_file:
        tree = ET.parse(in_file)
        root = tree.getroot()

    for obj in root.iter("object"):
        geometry = None
        obj_cls = None
        tags = []

        for element in obj:
            field_name, value = element.tag, element.text
            if field_name in ["name", "class"]:
                cls_name = bbox_classes_map.get(value, value)
                if renamed_classes and cls_name in renamed_classes:
                    cls_name = renamed_classes[cls_name]
                obj_cls = meta.obj_classes.get(cls_name)
                if obj_cls is None:
                    logger.warning(f"Class {cls_name} is not found in meta. Skipping.")
                    continue
            elif field_name == "bndbox":
                bbox_coords = [
                    float(element.find(x).text) for x in ("ymin", "xmin", "ymax", "xmax")
                ]
                geometry = Rectangle(*bbox_coords)
            elif field_name not in DEFAULT_OBJECT_FIELDS:
                tag_name = field_name
                if renamed_tags and tag_name in renamed_tags:
                    tag_name = renamed_tags[tag_name]
                tag_meta = meta.get_tag_meta(tag_name)
                if tag_meta is None:
                    logger.warning(f"Tag meta for '{field_name}' is not found in meta. Skipping.")
                    continue

                if tag_meta.value_type == TagValueType.ANY_STRING:
                    if not isinstance(value, str):
                        value = str(value)
                    tags.append(Tag(tag_meta, value))
                elif tag_meta.value_type == TagValueType.NONE:
                    # check if value is numeric
                    try:
                        int(value)
                    except ValueError:
                        logger.warning(f"Tag value is not numeric: '{value}'. Skipping.")
                        continue
                    if int(value) == 1:
                        tags.append(Tag(tag_meta))
                    else:
                        logger.debug("Tag with value '0' not added to labels.")
                elif tag_meta.value_type == TagValueType.ONEOF_STRING:
                    if value not in tag_meta.possible_values:
                        logger.warning(
                            f"Value '{value}' for tag '{tag_name}' is not in possible values: {tag_meta.possible_values}. Skipping."
                        )
                        continue
                    tags.append(Tag(tag_meta, value))

        if geometry is None or obj_cls is None:
            continue
        labels.append(Label(geometry, obj_cls, tags))
    labels = validate_image_bounds(labels, img_rect)

    return labels


def update_meta_from_xml(
    xml_path: str,
    meta: ProjectMeta,
    existing_cls_names: set,
    bbox_classes_map: dict,
    tags_to_values: Dict[str, set],
) -> ProjectMeta:
    import xml.etree.ElementTree as ET

    with open(xml_path, "r") as in_file:
        tree = ET.parse(in_file)
        root = tree.getroot()

    for obj in root.iter("object"):
        for element in obj:
            field_name = element.tag
            if field_name in ["name", "class"]:
                class_name = element.text
                original_class_name = class_name
                obj_cls = meta.obj_classes.get(class_name)
                if obj_cls is None:
                    obj_cls = ObjClass(name=class_name, geometry_type=Rectangle)
                    meta = meta.add_obj_class(obj_cls)
                    existing_cls_names.add(class_name)
                    continue
                elif obj_cls.geometry_type == Rectangle:
                    continue
                class_name = class_name + "_bbox"
                obj_cls = meta.obj_classes.get(class_name)
                if obj_cls is None:
                    obj_cls = ObjClass(name=class_name, geometry_type=Rectangle)
                    meta = meta.add_obj_class(obj_cls)
                    existing_cls_names.add(class_name)
                elif obj_cls.geometry_type == Rectangle:
                    pass
                else:
                    class_name = generate_free_name(
                        existing_cls_names, class_name, extend_used_names=True
                    )
                    obj_cls = ObjClass(name=class_name, geometry_type=Rectangle)
                    meta = meta.add_obj_class(obj_cls)
                bbox_classes_map[original_class_name] = class_name
            elif field_name not in DEFAULT_OBJECT_FIELDS:
                value = element.text
                tags_to_values[field_name].add(value)

    return meta


def sly_ann_to_pascal_voc(ann: Annotation, image_name: str) -> Tuple[dict]:
    """
    Convert Supervisely annotation to Pascal VOC format annotation.

    :param ann: Supervisely annotation.
    :type ann: :class:`Annotation<supervisely.annotation.annotation.Annotation>`
    :param image_name: Image name.
    :type image_name: :class:`str`
    :return: Tuple with xml tree and instance and class masks in PIL.Image format.
    :rtype: :class:`Tuple`

    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.convert.image.pascal_voc.pascal_voc_helper import sly_ann_to_pascal_voc

        ann = sly.Annotation.from_json(ann_json, meta)
        xml_tree, instance_mask, class_mask = sly_ann_to_pascal_voc(ann, image_name)
    """

    def from_ann_to_instance_mask(ann: Annotation, contour_thickness: int = 3):
        mask = np.zeros((ann.img_size[0], ann.img_size[1], 3), dtype=np.uint8)
        for label in ann.labels:
            if label.obj_class.geometry_type == Rectangle:
                continue

            if label.obj_class.name == "neutral":
                label.geometry.draw(mask, default_classes_colors["neutral"])
                continue

            label.geometry.draw_contour(mask, default_classes_colors["neutral"], contour_thickness)
            label.geometry.draw(mask, label.obj_class.color)

        res_mask = Image.fromarray(mask)
        res_mask = res_mask.convert("P", palette=Image.ADAPTIVE)  # pylint: disable=no-member
        return res_mask

    def from_ann_to_class_mask(ann: Annotation, contour_thickness: int = 3):
        exist_colors = [[0, 0, 0], default_classes_colors["neutral"]]
        mask = np.zeros((ann.img_size[0], ann.img_size[1], 3), dtype=np.uint8)
        for label in ann.labels:
            if label.obj_class.geometry_type == Rectangle:
                continue

            if label.obj_class.name == "neutral":
                label.geometry.draw(mask, default_classes_colors["neutral"])
                continue

            new_color = generate_rgb(exist_colors)
            exist_colors.append(new_color)
            label.geometry.draw_contour(mask, default_classes_colors["neutral"], contour_thickness)
            label.geometry.draw(mask, new_color)

        res_mask = Image.fromarray(mask)
        res_mask = res_mask.convert("P", palette=Image.ADAPTIVE)  # pylint: disable=no-member
        return res_mask

    def from_ann_to_xml(ann: Annotation, image_name: str):
        import lxml.etree as ET  # pylint: disable=import-error

        xml_root = ET.Element("annotation")

        ET.SubElement(xml_root, "folder").text = f"VOC"
        ET.SubElement(xml_root, "filename").text = image_name

        xml_root_source = ET.SubElement(xml_root, "source")
        ET.SubElement(xml_root_source, "database").text = ""

        ET.SubElement(xml_root_source, "annotation").text = "PASCAL VOC"
        ET.SubElement(xml_root_source, "image").text = ""

        xml_root_size = ET.SubElement(xml_root, "size")
        ET.SubElement(xml_root_size, "width").text = str(ann.img_size[1])
        ET.SubElement(xml_root_size, "height").text = str(ann.img_size[0])
        ET.SubElement(xml_root_size, "depth").text = "3"

        ET.SubElement(xml_root, "segmented").text = "1" if len(ann.labels) > 0 else "0"

        for label in ann.labels:
            if label.obj_class.name == "neutral":
                continue

            bitmap_to_bbox = label.geometry.to_bbox()

            xml_ann_obj = ET.SubElement(xml_root, "object")
            ET.SubElement(xml_ann_obj, "name").text = label.obj_class.name
            ET.SubElement(xml_ann_obj, "pose").text = "Unspecified"
            ET.SubElement(xml_ann_obj, "truncated").text = "0"
            ET.SubElement(xml_ann_obj, "difficult").text = "0"

            xml_ann_obj_bndbox = ET.SubElement(xml_ann_obj, "bndbox")
            ET.SubElement(xml_ann_obj_bndbox, "xmin").text = str(bitmap_to_bbox.left)
            ET.SubElement(xml_ann_obj_bndbox, "ymin").text = str(bitmap_to_bbox.top)
            ET.SubElement(xml_ann_obj_bndbox, "xmax").text = str(bitmap_to_bbox.right)
            ET.SubElement(xml_ann_obj_bndbox, "ymax").text = str(bitmap_to_bbox.bottom)

        tree = ET.ElementTree(xml_root)
        return tree

    pascal_ann = from_ann_to_xml(ann, image_name)
    instance_mask = from_ann_to_instance_mask(ann)
    class_mask = from_ann_to_class_mask(ann)
    return pascal_ann, instance_mask, class_mask


def sly_ds_to_pascal_voc(
    dataset: Dataset,
    meta: ProjectMeta,
    dest_dir: Optional[str] = None,
    train_val_split_coef: float = 0.8,
    log_progress: bool = False,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
) -> None:
    """
    Convert Supervisely dataset to Pascal VOC format.

    :param meta: Project meta information.
    :type meta: :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>`
    :param dest_dir: Destination directory.
    :type dest_dir: :class:`str`, optional
    :param train_val_split_coef: Coefficient for splitting images into train and validation sets.
    :type train_val_split_coef: :class:`float`, optional
    :param log_progress: If True, log progress.
    :type log_progress: :class:`str`, optional
    :param progress_cb: Progress callback.
    :type progress_cb: :class:`Callable`, optional
    :return: None
    :rtype: NoneType

    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.convert.image.pascal_voc.pascal_voc_helper import sly_ds_to_pascal_voc

        project_path = "/home/admin/work/supervisely/projects/lemons_annotated"
        project = sly.Project(project_path, sly.OpenMode.READ)

        for ds in project.datasets:
            dest_dir = "/home/admin/work/supervisely/projects/lemons_annotated_pascal_voc"
            sly_ds_to_pascal_voc(ds, project.meta, dest_dir=dest_dir)
    """
    import lxml.etree as ET  # pylint: disable=import-error

    def write_main_set(
        is_trainval: int,
        images_stats: dict,
        meta: ProjectMeta,
        result_main_sets_dir: str,
        result_segmentation_sets_dir: str,
    ):
        res_files = ["trainval.txt", "train.txt", "val.txt"]
        for file in os.listdir(result_segmentation_sets_dir):
            if file in res_files:
                shutil.copyfile(
                    os.path.join(result_segmentation_sets_dir, file),
                    os.path.join(result_main_sets_dir, file),
                )

        train_imgs = [i for i in images_stats if i["dataset"] == TRAIN_TAG_NAME]
        val_imgs = [i for i in images_stats if i["dataset"] == VAL_TAG_NAME]

        write_objs = [
            {"suffix": "trainval", "imgs": images_stats},
            {"suffix": "train", "imgs": train_imgs},
            {"suffix": "val", "imgs": val_imgs},
        ]

        if is_trainval == 1:
            trainval_imgs = [
                i for i in images_stats if i["dataset"] == TRAIN_TAG_NAME + VAL_TAG_NAME
            ]
            write_objs[0] = {"suffix": "trainval", "imgs": trainval_imgs}

        for obj_cls in meta.obj_classes:
            if obj_cls.geometry_type not in SUPPORTED_GEOMETRY_TYPES:
                continue
            if obj_cls.name == "neutral":
                continue
            for o in write_objs:
                with open(
                    os.path.join(result_main_sets_dir, f'{obj_cls.name}_{o["suffix"]}.txt'), "a"
                ) as f:
                    for img_stats in o["imgs"]:
                        v = "1" if obj_cls.name in img_stats["classes"] else "-1"
                        f.write(f'{img_stats["name"]} {v}\n')

    def write_segm_set(is_trainval: int, images_stats: dict, result_imgsets_dir: str):
        with open(os.path.join(result_imgsets_dir, "trainval.txt"), "a") as f:
            if is_trainval == 1:
                f.writelines(
                    i["name"] + "\n"
                    for i in images_stats
                    if i["dataset"] == TRAIN_TAG_NAME + VAL_TAG_NAME
                )
            else:
                f.writelines(i["name"] + "\n" for i in images_stats)
        with open(os.path.join(result_imgsets_dir, "train.txt"), "a") as f:
            f.writelines(i["name"] + "\n" for i in images_stats if i["dataset"] == TRAIN_TAG_NAME)
        with open(os.path.join(result_imgsets_dir, "val.txt"), "a") as f:
            f.writelines(i["name"] + "\n" for i in images_stats if i["dataset"] == VAL_TAG_NAME)

    if progress_cb is not None:
        log_progress = False

    if log_progress:
        progress_cb = tqdm_sly(
            desc=f"Converting dataset '{dataset.short_name}' to Pascal VOC format",
            total=len(dataset),
        )

    logger.info(f"Processing dataset: '{dataset.name}'")

    # Prepare Pascal VOC root directory
    if dest_dir is None:
        dest_dir = str(Path(dataset.path).parent / "pascal_voc")

    pascal_root_path = os.path.join(dest_dir, "VOCdevkit", "VOC")
    result_images_dir = os.path.join(pascal_root_path, "JPEGImages")
    result_ann_dir = os.path.join(pascal_root_path, "Annotations")
    result_obj_dir = os.path.join(pascal_root_path, "SegmentationObject")
    result_class_dir = os.path.join(pascal_root_path, "SegmentationClass")
    result_image_sets_dir = os.path.join(pascal_root_path, "ImageSets")
    result_segmentation_sets_dir = os.path.join(result_image_sets_dir, "Segmentation")
    result_main_sets_dir = os.path.join(result_image_sets_dir, "Main")
    result_colors_file_path = os.path.join(pascal_root_path, "colors.txt")

    # Create directories if not exist
    os.makedirs(result_images_dir, exist_ok=True)
    os.makedirs(result_ann_dir, exist_ok=True)
    os.makedirs(result_obj_dir, exist_ok=True)
    os.makedirs(result_class_dir, exist_ok=True)
    os.makedirs(result_image_sets_dir, exist_ok=True)
    os.makedirs(result_segmentation_sets_dir, exist_ok=True)
    os.makedirs(result_main_sets_dir, exist_ok=True)

    # Create colors.txt file
    if not file_exists(result_colors_file_path):
        with open(result_colors_file_path, "w") as f:
            f.write(
                f"neutral {default_classes_colors['neutral'][0]} {default_classes_colors['neutral'][1]} {default_classes_colors['neutral'][2]}\n"
            )

    image_stats = []
    classes_colors = {}
    for item_name, img_path, ann_path in dataset.items():
        # Assign unique name to avoid conflicts
        unique_name = f"{dataset.name}_{get_file_name(item_name)}"

        # Load annotation
        ann = Annotation.from_json(load_json_file(ann_path), meta)
        pascal_ann, instance_mask, class_mask = sly_ann_to_pascal_voc(ann, unique_name)

        # Write ann
        ann_path = os.path.join(result_ann_dir, f"{unique_name}.xml")
        ET.indent(pascal_ann, space="    ")
        pascal_ann.write(ann_path, pretty_print=True)

        # Save instance mask
        instance_mask_path = os.path.join(
            result_obj_dir, f"{unique_name}_instance{MASKS_EXTENSION}"
        )
        instance_mask.save(instance_mask_path)

        # Save class mask
        class_mask_path = os.path.join(result_class_dir, f"{unique_name}_class{MASKS_EXTENSION}")
        class_mask.save(class_mask_path)

        # Save original image
        img_ext = get_file_ext(img_path)
        if img_ext not in VALID_IMG_EXT:
            jpg_name = f"{unique_name}.jpg"
            jpg_image_path = os.path.join(result_images_dir, jpg_name)
            img = Image.open(img_path)
            img.save(jpg_image_path, "JPEG")
        else:
            jpg_name = f"{unique_name}{img_ext}"
            jpg_image_path = os.path.join(result_images_dir, jpg_name)
            shutil.copyfile(img_path, jpg_image_path)

        # Update stats
        cur_img_stats = {"classes": set(), "dataset": None, "name": jpg_name}
        image_stats.append(cur_img_stats)

        # Get classes colors
        for label in ann.labels:
            cur_img_stats["classes"].add(label.obj_class.name)
            classes_colors[label.obj_class.name] = tuple(label.obj_class.color)

        if log_progress:
            progress_cb.update(1)

    # Update colors.txt file
    classes_colors = OrderedDict((sorted(classes_colors.items(), key=lambda t: t[0])))
    with open(result_colors_file_path, "a") as cc:
        for k in classes_colors.keys():
            if k == "neutral":
                continue
            cc.write(f"{k} {classes_colors[k][0]} {classes_colors[k][1]} {classes_colors[k][2]}\n")

    # Create splits
    imgs_to_split = [i for i in image_stats if i["dataset"] is None]
    train_len = int(len(imgs_to_split) * train_val_split_coef)

    for img_stat in imgs_to_split[:train_len]:
        img_stat["dataset"] = TRAIN_TAG_NAME
    for img_stat in imgs_to_split[train_len:]:
        img_stat["dataset"] = VAL_TAG_NAME

    is_trainval = 0
    write_segm_set(is_trainval, image_stats, result_segmentation_sets_dir)
    write_main_set(
        is_trainval, image_stats, meta, result_main_sets_dir, result_segmentation_sets_dir
    )


def sly_project_to_pascal_voc(
    project: Union[Project, str],
    dest_dir: Optional[str] = None,
    train_val_split_coef: float = 0.8,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
) -> None:
    """
    Convert Supervisely project to Pascal VOC format.

    :param dest_dir: Destination directory.
    :type dest_dir: :class:`str`, optional
    :param train_val_split_coef: Coefficient for splitting images into train and validation sets.
    :type train_val_split_coef: :class:`float`, optional
    :param log_progress: Show uploading progress bar.
    :type log_progress: :class:`bool`
    :param progress_cb: Function for tracking conversion progress (for all items in the project).
    :type progress_cb: callable, optional
    :return: None
    :rtype: NoneType

    :Usage example:

    .. code-block:: python

        import supervisely as sly

        # Local folder with Project
        project_directory = "/home/admin/work/supervisely/source/project"

        # Convert Project to Pascal VOC format
        sly.Project(project_directory).to_pascal_voc(log_progress=True)
    """
    if isinstance(project, str):
        project = Project(project, mode=OpenMode.READ)

    if dest_dir is None:
        dest_dir = project.directory

    Path(dest_dir).mkdir(parents=True, exist_ok=True)

    if progress_cb is not None:
        log_progress = False

    if log_progress:
        progress_cb = tqdm_sly(
            desc="Converting Supervisely project to Pascal VOC format", total=project.total_items
        )

    for dataset in project.datasets:
        dataset: Dataset
        dataset.to_pascal_voc(
            meta=project.meta,
            dest_dir=dest_dir,
            train_val_split_coef=train_val_split_coef,
            log_progress=log_progress,
            progress_cb=progress_cb,
        )
        logger.info(f"Dataset '{dataset.short_name}' has been converted to Pascal VOC format.")
    logger.info(f"Project '{project.name}' has been converted to Pascal VOC format.")


def to_pascal_voc(
    input_data: Union[Project, Dataset, str],
    dest_dir: Optional[str] = None,
    meta: Optional[ProjectMeta] = None,
    train_val_split_coef: float = 0.8,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
) -> None:
    """
    Universal function to convert Supervisely project or dataset to Pascal VOC format.
    Note:
        - For better compatibility, please pass named arguments explicitly. Otherwise, the function may not work as expected.
            You can use the dedicated functions for each data type:

                - :func:`sly.convert.sly_project_to_pascal_voc()`
                - :func:`sly.convert.sly_ds_to_pascal_voc()`

        - If the input_data is a Project, the dest_dir parameters are required.
        - If the input_data is a Dataset, the meta and dest_dir parameters are required.

    :param input_data: Input data to convert (Project, Dataset, or path to the project/dataset directory).
    :type input_data: :class:`Project<supervisely.project.project.Project>`, :class:`Dataset<supervisely.dataset.dataset.Dataset>`, or :class:`str`
    :param dest_dir: Destination directory.
    :type dest_dir: :class:`str`, optional
    :param meta: Project meta information (required for Dataset conversion).
    :type meta: :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>`, optional
    :param train_val_split_coef: Coefficient for splitting images into train and validation sets.
    :type train_val_split_coef: :class:`float`, optional
    :param log_progress: Show uploading progress bar.
    :type log_progress: :class:`bool`
    :param progress_cb: Function for tracking conversion progress (for all items in the project).
    :type progress_cb: callable, optional
    :return: None
    :rtype: NoneType

    :Usage example:

    .. code-block:: python

        import supervisely as sly

        # Local folder with Project
        project_directory = "/home/admin/work/supervisely/source/project"
        project_fs = sly.Project(project_directory, sly.OpenMode.READ)

        # Convert Project to Pascal VOC format
        sly.convert.to_pascal_voc(project_directory, dest_dir="./pascal_voc")
        # or
        sly.convert.to_pascal_voc(project_fs, dest_dir="./pascal_voc")

        # Convert Dataset to Pascal VOC format
        dataset: sly.Dataset = project_fs.datasets.get("dataset_name")
        sly.convert.to_pascal_voc(dataset, dest_dir="./pascal_voc")
    """
    if isinstance(input_data, str):
        try:
            input_data = Project(input_data, mode=OpenMode.READ)
        except Exception as e:
            try:
                input_data = Dataset(input_data, mode=OpenMode.READ)
            except Exception as e:
                raise ValueError("Please check the path or the input data.")

    if isinstance(input_data, (Project, str)):
        return sly_project_to_pascal_voc(
            project=input_data,
            dest_dir=dest_dir,
            train_val_split_coef=train_val_split_coef,
            log_progress=log_progress,
            progress_cb=progress_cb,
        )
    elif isinstance(input_data, Dataset):
        return sly_ds_to_pascal_voc(
            dataset=input_data,
            meta=meta,
            dest_dir=dest_dir,
            train_val_split_coef=train_val_split_coef,
            log_progress=log_progress,
            progress_cb=progress_cb,
        )
    else:
        raise ValueError(f"Unsupported input data type: {type(input_data)}")



from typing import Dict, List, Literal, Optional

from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets import Widget


class Modal(Widget):
    """Modal is a widget that displays content in a modal overlay window with a close button.
    It can contain any other widgets, similar to Container, and provides programmatic control
    to show/hide the modal.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/layouts-and-containers/modal>`_
        (including screenshots and examples).

    :param title: Modal window title
    :type title: str
    :param widgets: List of widgets to be displayed inside the modal
    :type widgets: Optional[List[Widget]]
    :param size: Modal size, one of: tiny, small, large, full
    :type size: Literal["tiny", "small", "large", "full"]
    :param widget_id: An identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
    .. code-block:: python

        from supervisely.app.widgets import Modal, Text, Button, Input, Container

        # Create widgets to show in modal
        input_widget = Input("Enter value")
        text_widget = Text("This is modal content")

        # Create modal with multiple widgets
        modal = Modal(
            title="My Modal Window",
            widgets=[text_widget, input_widget],
            size="small"
        )

        # Show modal programmatically
        modal.show()

        # Hide modal programmatically
        modal.hide()

        # Alternative methods
        modal.show_modal()
        modal.close_modal()
    """

    class Routes:
        ON_CLOSE = "close_cb"
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        title: Optional[str] = "",
        widgets: Optional[List[Widget]] = None,
        size: Optional[Literal["tiny", "small", "large", "full"]] = "small",
        widget_id: Optional[str] = None,
    ):
        self._title = title
        self._widgets = widgets if widgets is not None else []
        self._size = size
        self._value_changed_handled = False
        super().__init__(widget_id=widget_id, file_path=__file__)

        server = self._sly_app.get_server()
        route = self.get_route_path(Modal.Routes.ON_CLOSE)

        @server.post(route)
        def _on_close():
            # Change visibility state to False when modal is closed on client side
            visible = StateJson()[self.widget_id]["visible"]
            if visible is True:
                StateJson()[self.widget_id]["visible"] = False
                # no need to call send_changes(), as it is already changed on client side

    def get_json_data(self) -> Dict[str, str]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - title: Modal title
            - size: Modal size, one of: tiny, small, large, full

        :return: Dictionary with widget data
        :rtype: Dict[str, str]
        """
        return {
            "title": self._title,
            "size": self._size,
        }

    def get_json_state(self) -> Dict[str, bool]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - visible: Modal visibility

        :return: Dictionary with widget state
        :rtype: Dict[str, bool]
        """
        return {
            "visible": False,
        }

    def show(self) -> None:
        """Shows the modal window."""
        StateJson()[self.widget_id]["visible"] = True
        StateJson().send_changes()

    def hide(self) -> None:
        """Hides the modal window."""
        StateJson()[self.widget_id]["visible"] = False
        StateJson().send_changes()

    def show_modal(self) -> None:
        """Shows the modal window. Alias for show() method."""
        self.show()

    def close_modal(self) -> None:
        """Closes the modal window. Alias for hide() method."""
        self.hide()

    def is_opened(self) -> bool:
        """Returns whether the modal is currently open.

        :return: True if modal is visible, False otherwise
        :rtype: bool
        """
        return StateJson()[self.widget_id]["visible"]

    def value_changed(self, func):
        """Decorator to handle modal visibility changes.
        The callback function will receive a boolean parameter: True when opened, False when closed.

        :param func: Callback function that takes a boolean parameter
        :type func: Callable[[bool], None]

        :Usage example:
        .. code-block:: python

            @modal.value_changed
            def on_modal_state_changed(is_opened):
                if is_opened:
                    print("Modal opened")
                else:
                    print("Modal closed")
        """
        route_path = self.get_route_path(Modal.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._value_changed_handled = True

        @server.post(route_path)
        def _value_changed():
            is_opened = StateJson()[self.widget_id]["visible"]
            func(is_opened)

        return _value_changed

    @property
    def title(self) -> str:
        """Returns modal title.

        :return: Modal title
        :rtype: str
        """
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        """Sets modal title.

        :param title: Modal title
        :type title: str
        """
        self._title = title
        DataJson()[self.widget_id]["title"] = title
        DataJson().send_changes()

    @property
    def widgets(self) -> List[Widget]:
        """Returns list of widgets inside the modal.

        :return: List of widgets
        :rtype: List[Widget]
        """
        return self._widgets

    @widgets.setter
    def widgets(self, widgets: List[Widget]) -> None:
        """Sets the list of widgets to display in the modal.
        Note: Changing widgets dynamically may require re-rendering.

        :param widgets: List of widgets
        :type widgets: List[Widget]
        """
        self._widgets = widgets



import typing
from os import PathLike
import jinja2


def create_env(directory: typing.Union[str, PathLike]) -> "jinja2.Environment":
    env_sly = jinja2.Environment(
        loader=jinja2.FileSystemLoader(directory),
        autoescape=True,
        variable_start_string="{{{",
        variable_end_string="}}}",
    )
    return env_sly



from typing import Any, Dict

from supervisely.app.widgets import Button, Card, Container, TagsTable, Text


class TagsSelector:
    title = "Tags Selector"
    description = "Select tags that will be used for inference"
    lock_message = "Select previous step to unlock"

    def __init__(self):
        # Init Step
        self.display_widgets = []
        # -------------------------------- #

        # Init Base Widgets
        self.validator_text = None
        self.button = None
        self.container = None
        self.card = None
        # -------------------------------- #

        # Init Step Widgets
        self.tags_table = None
        # -------------------------------- #

        # Tags
        self.tags_table = TagsTable()
        self.tags_table.hide()
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.tags_table])
        # ----------------------------------- #

        # Base Widgets
        self.validator_text = Text("")
        self.validator_text.hide()
        self.button = Button("Select")
        self.display_widgets.extend([self.validator_text, self.button])
        # -------------------------------- #

        # Card Layout
        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
        )
        # -------------------------------- #

    @property
    def widgets_to_disable(self) -> list:
        return [self.tags_table]

    def load_from_json(self, data: Dict[str, Any]) -> None:
        if "tags" in data:
            self.set_tags(data["tags"])

    def get_selected_tags(self) -> list:
        return self.tags_table.get_selected_tags()

    def set_tags(self, tags) -> None:
        self.tags_table.select_tags(tags)

    def select_all_tags(self) -> None:
        self.tags_table.select_all()

    def get_settings(self) -> Dict[str, Any]:
        return {"tags": self.get_selected_tags()}

    def validate_step(self) -> bool:
        if self.tags_table.is_hidden():
            return True

        self.validator_text.hide()

        project_tags = self.tags_table.project_meta.tag_metas
        if len(project_tags) == 0:
            self.validator_text.set(text="Project has no tags", status="error")
            self.validator_text.show()
            return False

        selected_tags = self.tags_table.get_selected_tags()
        table_data = self.tags_table._table_data
        empty_tags = [
            row[0]["data"]
            for row in table_data
            if row[0]["data"] in selected_tags and row[2]["data"] == 0 and row[3]["data"] == 0
        ]

        n_tags = len(selected_tags)
        if n_tags == 0:
            message = "Please select at least one tag"
            status = "error"
        else:
            tag_text = "tag" if n_tags == 1 else "tags"
            message = f"Selected {n_tags} {tag_text}"
            status = "success"
            if empty_tags:
                intersections = set(selected_tags).intersection(empty_tags)
                if intersections:
                    tag_text = "tag" if len(intersections) == 1 else "tags"
                    message += (
                        f". Selected {tag_text} have no annotations: {', '.join(intersections)}"
                    )
                    status = "warning"

        self.validator_text.set(text=message, status=status)
        self.validator_text.show()
        return n_tags > 0
